"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import inspect
import logging
import base64
import traceback
from concurrent.futures import Future
from collections import defaultdict
import typing


from tornado import concurrent
from typing import Tuple

from inmanta.agent.io import get_io
from inmanta import protocol, resources, const, data
from inmanta.agent.cache import AgentCache
import uuid
from typing import Dict

LOGGER = logging.getLogger(__name__)


class provider(object):  # noqa: N801
    """
        A decorator that registers a new handler.

        :param resource_type: The type of the resource this handler provides an implementation for.
                              For example, :inmanta:entity:`std::File`
        :param name: A name to reference this provider.
    """

    def __init__(self, resource_type: str, name: str) -> None:
        self._resource_type = resource_type
        self._name = name

    def __call__(self, function):
        """
            The wrapping
        """
        Commander.add_provider(self._resource_type, self._name, function)
        return function


class SkipResource(Exception):
    """
        A handler should raise this exception when a resource should be skipped. The resource will be marked as skipped
        instead of failed.
    """


class ResourcePurged(Exception):
    """
        If the :func:`~inmanta.agent.handler.CRUDHandler.read_resource` method raises this exception, the agent will
        mark the current state of the resource as purged.
    """


def cache(f=None, ignore: typing.List[str]=[], timeout: int=5000, for_version: bool=True, cache_none: bool=True,
          cacheNone: bool=True):  # noqa: N803
    """
        decorator for methods in resource handlers to provide caching

        this decorator works similar to memoization:
        when the decorate method is called, its return value is cached,
        for subsequent calls, the cached value is used instead of the actual value

        The name of the method + the arguments of the method form the cache key

        If an argument named version is present and for_version is True,
        the cache entry is flushed after this version has been deployed
        If an argument named resource is present,
        it is assumed to be a resource and its ID is used, without the version information

        :param timeout: the number of second this cache entry should live
        :param for_version: if true, this value is evicted from the cache when this deploy is ready
        :param ignore: a list of argument names that should not be part of the cache key
        :param cache_none: cache returned none values
    """

    def actual(f):
        myignore = set(ignore)
        sig = inspect.signature(f)
        myargs = list(sig.parameters.keys())[1:]

        def wrapper(self, *args, **kwds):

            kwds.update(dict(zip(myargs, args)))

            def bound(**kwds):
                return f(self, **kwds)

            cache_none = cacheNone
            return self.cache.get_or_else(f.__name__, bound, for_version, timeout, myignore, cache_none, **kwds)

        return wrapper

    if f is None:
        return actual
    else:
        return actual(f)


class HandlerContext(object):
    """
        Context passed to handler methods for state related "things"
    """

    def __init__(self, resource, dry_run=False, action_id=None, logger=None):
        self._resource = resource
        self._dry_run = dry_run
        self._cache = {}

        self._purged = False
        self._updated = False
        self._created = False
        self._change = const.Change.nochange

        self._changes: Dict[str, Dict[str, str]] = {}

        if action_id is None:
            action_id = uuid.uuid4()
        self._action_id = action_id
        self._status = None
        self._logs = []
        if logger is None:
            self.logger = LOGGER
        else:
            self.logger = logger

    @property
    def action_id(self) -> uuid.UUID:
        return self._action_id

    @property
    def status(self) -> const.ResourceState:
        return self._status

    @property
    def logs(self) -> list:
        return self._logs

    def set_status(self, status: const.ResourceState) -> None:
        """
            Set the status of the handler operation.
        """
        self._status = status

    def is_dry_run(self):
        """
            Is this a dryrun?
        """
        return self._dry_run

    def get(self, name):
        return self._cache[name]

    def contains(self, key):
        return key in self._cache

    def set(self, name, value):
        self._cache[name] = value

    def set_created(self):
        self._created = True
        if self._change is None or self._change.value < const.Change.created.value:
            self._change = const.Change.created

    def set_purged(self):
        self._purged = True
        if self._change is None or self._change.value < const.Change.purged.value:
            self._change = const.Change.purged

    def set_updated(self):
        self._updated = True
        if self._change is None or self._change.value < const.Change.updated.value:
            self._change = const.Change.updated

    @property
    def changed(self):
        return self._created or self._updated or self._purged

    @property
    def change(self) -> const.Change:
        return self._change

    def add_change(self, name: str, desired: typing.Any, current: typing.Any=None) -> None:
        """
            Report a change of a field. This field is added to the set of updated fields

            :param name: The name of the field that was updated
            :param desired: The desired value to which the field was updated (or should be updated)
            :param current: The value of the field before it was updated
        """
        self._changes[name] = {"current": current, "desired": desired}

    def add_changes(self, **kwargs):
        """
            Report a list of changes at once as kwargs

            :param key: The name of the field that was updated. This field is also added to the set of updated fields
            :param value: The desired value of the field.

            To report the previous value of the field, use the add_change method
        """
        for field, value in kwargs.items():
            self._changes[field] = {"desired": value}

    def fields_updated(self, fields):
        """
            Report that fields have been updated
        """
        for field in fields:
            if field not in self._changes:
                self._changes[fields] = {}

    def update_changes(self, changes: dict):
        """
            Update the changes list with changes

            :param changes: This should be a dict with a value a dict containing "current" and "desired" keys
        """
        self._changes.update(changes)

    @property
    def changes(self) -> Dict[str, Dict[str, str]]:
        return self._changes

    def log_msg(self, level: int, msg: str, args: list, kwargs: dict) -> dict:
        if len(args) > 0:
            raise Exception("Args not supported")
        if "exc_info" in kwargs:
            exc_info = kwargs["exc_info"]
            kwargs["traceback"] = traceback.format_exc()
        else:
            exc_info = False
        log = data.LogLine.log(level, msg, **kwargs)
        LOGGER.log(level, log._data["msg"], exc_info=exc_info)
        self._logs.append(log)

    def debug(self, msg: str, *args, **kwargs) -> None:
        """
        Log 'msg % args' with severity 'DEBUG'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        ``logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)``
        """
        self.log_msg(logging.DEBUG, msg, args, kwargs)

    def info(self, msg: str, *args, **kwargs) -> None:
        """
        Log 'msg % args' with severity 'INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        ``logger.info("Houston, we have a %s", "interesting problem", exc_info=1)``
        """
        self.log_msg(logging.INFO, msg, args, kwargs)

    def warning(self, msg: str, *args, **kwargs) -> None:
        """
        Log 'msg % args' with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        ``logger.warning("Houston, we have a %s", "bit of a problem", exc_info=1)``
        """
        self.log_msg(logging.WARNING, msg, args, kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        """
        Log 'msg % args' with severity 'ERROR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        ``logger.error("Houston, we have a %s", "major problem", exc_info=1)``
        """
        self.log_msg(logging.ERROR, msg, args, kwargs)

    def exception(self, msg: str, *args, exc_info=True, **kwargs) -> None:
        """
        Convenience method for logging an ERROR with exception information.
        """
        self.error(msg, *args, exc_info=exc_info, **kwargs)

    def critical(self, msg: str, *args, **kwargs) -> None:
        """
        Log 'msg % args' with severity 'CRITICAL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        ``logger.critical("Houston, we have a %s", "major disaster", exc_info=1)``
        """
        self.log_msg(logging.CRITICAL, msg, args, kwargs)


class ResourceHandler(object):
    """
        A baseclass for classes that handle resources. New handler are registered with the
        :func:`~inmanta.agent.handler.provider` decorator.

        The implementation of a handler should use the ``self._io`` instance to execute io operations. This io objects
        makes abstraction of local or remote operations. See :class:`~inmanta.agent.io.local.LocalIO` for the available
        operations.

        :param agent: The agent that is executing this handler.
        :param io: The io object to use.
    """

    def __init__(self, agent, io=None) -> None:
        self._agent = agent

        if io is None:
            raise Exception("Unsupported: no resource mgmt in RH")
        else:
            self._io = io

        self._client = None
        # explicit ioloop reference, as we don't want the ioloop for the current thread, but the one for the agent
        self._ioloop = agent.process._io_loop

    def run_sync(self, func: typing.Callable) -> typing.Any:
        """
            Run a the given async function on the ioloop of the agent. It will block the current thread until the future
            resolves.

            :param func: A function that returns a yieldable future.
            :return: The result of the async function.
        """
        f = Future()

        def run():
            try:
                result = func()
                if result is not None:
                    from tornado.gen import convert_yielded
                    result = convert_yielded(result)
                    concurrent.chain_future(result, f)
            except Exception as e:
                f.set_exception(e)

        self._ioloop.add_callback(run)

        return f.result()

    def set_cache(self, cache: AgentCache) -> None:
        self.cache = cache

    def get_client(self) -> protocol.SessionClient:
        """
            Get the client instance that identifies itself with the agent session.

            :return: A client that is associated with the session of the agent that executes this handler.
        """
        if self._client is None:
            self._client = protocol.SessionClient("agent", self._agent.sessionid)
        return self._client

    def process_events(self, ctx: HandlerContext, resource: resources.Resource, events: dict) -> None:
        """
            Process events generated by changes to required resources. Override this method to process events in a handler.

            The default implementation provides the reload mechanism. It will call do_reload when the handler can_reload() and
            if at least one of the dependents have successfully deployed and there were changes. Make sure to call this method
            from a subclass if the reload behaviour is required.

            This method is called for all dependents of the given resource (inverse of the requires relationship)
            that have send_event set to true and for which a deploy was started. These are the only conditions, even if all
            dependents have failed or no changes were deployed. It is up to the handler to filter out irrelevant events.

            In case of partial deployments (e.g. incremental deploy), only those resources that are being deployed will produce
            an event. I.e. it is possible to receive less events then expected.

            In case of failure of agent, server or the system being managed, delivery of events can not be guaranteed.
            Update events can be lost unrecoverably in case the agent or server fails after the update was performed, but before
            the event was emitted. In the current implementation, start of a new deploy while another is in progress can also
            causes updates to be lost.

            However, while event delivery can not be guaranteed, convergence to the desired state can be reliably detected.
            If the record of the convergence is lost, it will be retried until it is recorded.
            For strong behavioral guarantees, it is better to rely on desired state than on events.

            Events are best used to accelerate convergence.
            For example, cross agent dependencies primarily make use of the deployment log on the server to determine if their
            dependencies are in their desired state. To speed up convergence, events are sent to notify other agents of relevant
            changes to resources they depend on.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to process the events for.
            :param dict: A dict with events of the resource the given resource requires. The keys of the dict are the resources.
                         Each value is a dict with the items status (const.ResourceState), changes (dict) and
                         change (const.Change).
        """

        # ctx.status == const.ResourceState.deployed is only true if
        # 1- this resource was deployed (i.e. this is not a run to process events of failed dependencies)
        # 2- this resource was deployed successfully

        if self.can_reload() and ctx.status == const.ResourceState.deployed:
            reload = False
            for res, result in events.items():
                ctx.debug("Processing changes of %(res)s", res=res, result=result)
                if result["status"] == const.ResourceState.deployed and len(result["changes"]) > 0:
                    reload = True
                    break

            if reload:
                self.do_reload(ctx, resource)

    def can_process_events(self) -> bool:
        """
            Can this handler process events? This is a more generic version of the reload mechanism.

            See the :py:func:`ResourceHandler.process_events` for more details about this mechanism.

            :return: Return true if this handler processes events.
        """
        return self.can_reload()

    def can_reload(self) -> bool:
        """
            Can this handler reload?

            :return: Return true if this handler needs to reload on requires changes.
        """
        return False

    def do_reload(self, ctx: HandlerContext, resource: resources.Resource) -> None:
        """
            Perform a reload of this resource.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to reload.
        """

    def pre(self, ctx: HandlerContext, resource: resources.Resource) -> None:
        """
            Method executed before a handler operation (Facts, dryrun, real deployment, ...) is executed. Override this method
            to run before an operation.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to query facts for.
        """

    def post(self, ctx: HandlerContext, resource: resources.Resource) -> None:
        """
            Method executed after an operation. Override this method to run after an operation.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to query facts for.
        """

    def close(self) -> None:
        pass

    def _diff(self, current: resources.Resource, desired: resources.Resource) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        """
            Calculate the diff between the current and desired resource state.

            :param current: The current state of the resource
            :param desired: The desired state of the resource
            :return: A dict with key the name of the field and value another dict with "current" and "desired" as keys for
                     fields that require changes.
        """
        changes = {}

        # check attributes
        for field in current.__class__.fields:
            current_value = getattr(current, field)
            desired_value = getattr(desired, field)

            if current_value != desired_value and desired_value is not None:
                changes[field] = {"current": current_value, "desired": desired_value}

        return changes

    def check_resource(self, ctx: HandlerContext, resource: resources.Resource) -> resources.Resource:
        """
            Check the current state of a resource

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to check the current state of.
            :return: A resource to represents the current state. Use the :func:`~inmanta.resources.Resource.clone` to create
                     clone of the given resource that can be modified.
        """
        raise NotImplementedError()

    def list_changes(self, ctx: HandlerContext, resource: resources.Resource) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        """
            Returns the changes required to bring the resource on this system in the state described in the resource entry.
            This method calls :func:`~inmanta.agent.handler.ResourceHandler.check_resource`

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to check the current state of.
            :return: A dict with key the name of the field and value another dict with "current" and "desired" as keys for
                     fields that require changes.
        """
        current = self.check_resource(ctx, resource)
        return self._diff(current, resource)

    def do_changes(self, ctx: HandlerContext, resource: resources.Resource, changes: dict) -> None:
        """
            Do the changes required to bring the resource on this system in the state of the given resource.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to check the current state of.
            :param changes: The changes that need to occur as reported by
                            :func:`~inmanta.agent.handler.ResourceHandler.list_changes`
        """
        raise NotImplementedError()

    def execute(self, ctx: HandlerContext, resource: resources.Resource, dry_run: bool=False) -> None:
        """
            Update the given resource. This method is called by the agent. Most handlers will not override this method
            and will only override :func:`~inmanta.agent.handler.ResourceHandler.check_resource`, optionally
            :func:`~inmanta.agent.handler.ResourceHandler.list_changes` and
            :func:`~inmanta.agent.handler.ResourceHandler.do_changes`

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to check the current state of.
            :param dry_run: True will only determine the required changes but will not execute them.
        """
        try:
            self.pre(ctx, resource)

            changes = self.list_changes(ctx, resource)
            ctx.update_changes(changes)

            if not dry_run:
                self.do_changes(ctx, resource, changes)
                ctx.set_status(const.ResourceState.deployed)

            else:
                ctx.set_status(const.ResourceState.dry)

        except SkipResource as e:
            ctx.set_status(const.ResourceState.skipped)
            ctx.warning(msg="Resource %(resource_id)s was skipped: %(reason)s", resource_id=resource.id, reason=e.args)

        except Exception as e:
            ctx.set_status(const.ResourceState.failed)
            ctx.exception("An error occurred during deployment of %(resource_id)s (exception: %(exception)s",
                          resource_id=resource.id, exception=repr(e))
        finally:
            try:
                self.post(ctx, resource)
            except Exception as e:
                ctx.exception("An error occurred after deployment of %(resource_id)s (exception: %(exception)s",
                              resource_id=resource.id, exception=repr(e))

    def facts(self, ctx: HandlerContext, resource: resources.Resource) -> dict:
        """
            Returns facts about this resource. Override this method to implement fact querying.
            :func:`~inmanta.agent.handler.ResourceHandler.pre` and :func:`~inmanta.agent.handler.ResourceHandler.post` are
            called before and after this method.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to query facts for.
            :return: A dict with fact names as keys and facts values.
        """
        return {}

    def check_facts(self, ctx: HandlerContext, resource: resources.Resource) -> dict:
        """
            This method is called by the agent to query for facts. It runs :func:`~inmanta.agent.handler.ResourceHandler.pre`
            and :func:`~inmanta.agent.handler.ResourceHandler.post`. This method calls
            :func:`~inmanta.agent.handler.ResourceHandler.facts` to do the actually querying.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to query facts for.
            :return: A dict with fact names as keys and facts values.
        """
        try:
            self.pre(ctx, resource)
            facts = self.facts(ctx, resource)
        finally:
            try:
                self.post(ctx, resource)
            except Exception as e:
                ctx.exception("An error occurred after getting facts about %(resource_id)s (exception: %(exception)s",
                              resource_id=resource.id, exception=repr(e))

        return facts

    def available(self, resource: resources.Resource) -> bool:
        """
            Returns true if this handler is available for the given resource

            :param resource: Is this handler available for the given resource?
            :return: Available or not?
        """
        return True

    def get_file(self, hash_id) -> bytes:
        """
            Retrieve a file from the fileserver identified with the given id. The convention is to use the sha1sum of the
            content to identify it.

            :param hash_id: The id of the content/file to retrieve from the server.
            :return: The content in the form of a bytestring or none is the content does not exist.
        """
        def call():
            return self.get_client().get_file(hash_id)

        result = self.run_sync(call)
        if result.code == 404:
            return None
        elif result.code == 200:
            return base64.b64decode(result.result["content"])
        else:
            raise Exception("An error occurred while retrieving file %s" % hash_id)

    def stat_file(self, hash_id: str) -> bool:
        """
            Check if a file exists on the server. This method does and async call to the server and blocks on the result.

            :param hash_id: The id of the file on the server. The convention is the use the sha1sum of the content as id.
            :return: True if the file is available on the server.
        """
        def call():
            return self.get_client().stat_file(hash_id)

        result = self.run_sync(call)
        return result.code == 200

    def upload_file(self, hash_id: str, content: bytes) -> None:
        """
            Upload a file to the server

            :param hash_id: The id to identify the content. The convention is to use the sha1sum of the content to identify it.
            :param content: A byte string with the content
        """
        def call():
            return self.get_client().upload_file(id=hash_id, content=base64.b64encode(content).decode("ascii"))

        try:
            self.run_sync(call)
        except Exception:
            raise Exception("Unable to upload file to the server.")


class CRUDHandler(ResourceHandler):
    """
        This handler base class requires CRUD methods to be implemented: create, read, update and delete. Such a handler
        only works on purgeable resources.
    """

    def read_resource(self, ctx: HandlerContext, resource: resources.PurgeableResource) -> None:
        """
            This method reads the current state of the resource. It provides a copy of the resource that should be deployed,
            the method implementation should modify the attributes of this resource to the current state.

            :param ctx: Context can be used to pass value discovered in the read method to the CUD methods. For example, the
                       id used in API calls
            :param resource: A clone of the desired resource state. The read method need to set values on this object.
            :raise SkipResource: Raise this exception when the handler should skip this resource
            :raise ResourcePurged: Raise this exception when the resource does not exist yet.
        """

    def create_resource(self, ctx: HandlerContext, resource: resources.PurgeableResource) -> None:
        """
            This method is called by the handler when the resource should be created.

            :param context: Context can be used to get values discovered in the read method. For example, the id used in API
                            calls. This context should also be used to let the handler know what changes were made to the
                            resource.
            :param resource: The desired resource state.
        """

    def delete_resource(self, ctx: HandlerContext, resource: resources.PurgeableResource) -> None:
        """
            This method is called by the handler when the resource should be deleted.

            :param ctx: Context can be used to get values discovered in the read method. For example, the id used in API
                        calls. This context should also be used to let the handler know what changes were made to the
                        resource.
            :param resource: The desired resource state.
        """

    def update_resource(self, ctx: HandlerContext, changes: dict, resource: resources.PurgeableResource) -> None:
        """
            This method is called by the handler when the resource should be updated.

            :param ctx: Context can be used to get values discovered in the read method. For example, the id used in API
                        calls. This context should also be used to let the handler know what changes were made to the
                        resource.
            :param changes: A map of resource attributes that should be changed. Each value is a tuple with the current and the
                            desired value.
            :param resource: The desired resource state.
        """

    def execute(self, ctx: HandlerContext, resource: resources.PurgeableResource, dry_run: bool=None) -> None:
        """
            Update the given resource. This method is called by the agent. Override the CRUD methods of this class.

            :param ctx: Context object to report changes and logs to the agent and server.
            :param resource: The resource to check the current state of.
            :param dry_run: True will only determine the required changes but will not execute them.
        """
        try:
            self.pre(ctx, resource)

            # current is clone, except for purged is set to false to prevent a bug that occurs often where the desired
            # state defines purged=true but the read_resource fails to set it to false if the resource does exist
            current = resource.clone(purged=False)
            changes = {}
            try:
                self.read_resource(ctx, current)
                changes = self._diff(current, resource)

            except ResourcePurged:
                if not resource.purged:
                    changes["purged"] = dict(desired=resource.purged, current=True)

            for field, values in changes.items():
                ctx.add_change(field, desired=values["desired"], current=values["current"])

            if not dry_run:
                if "purged" in changes:
                    if not changes["purged"]["desired"]:
                        self.create_resource(ctx, resource)
                    else:
                        self.delete_resource(ctx, resource)

                elif len(changes) > 0:
                    self.update_resource(ctx, changes, resource)

                ctx.set_status(const.ResourceState.deployed)
            else:
                ctx.set_status(const.ResourceState.dry)

        except SkipResource as e:
            ctx.set_status(const.ResourceState.skipped)
            ctx.warning(msg="Resource %(resource_id)s was skipped: %(reason)s", resource_id=resource.id, reason=e.args)

        except Exception as e:
            ctx.set_status(const.ResourceState.failed)
            ctx.exception("An error occurred during deployment of %(resource_id)s (exception: %(exception)s)",
                          resource_id=resource.id, exception=repr(e), traceback=traceback.format_exc())
        finally:
            try:
                self.post(ctx, resource)
            except Exception as e:
                ctx.exception("An error occurred after deployment of %(resource_id)s (exception: %(exception)s",
                              resource_id=resource.id, exception=repr(e))


class Commander(object):
    """
        This class handles commands
    """
    __command_functions = defaultdict(dict)
    __handlers = []
    __handler_cache = {}

    @classmethod
    def get_handlers(cls):
        return cls.__command_functions

    @classmethod
    def reset(cls):
        cls.__command_functions = defaultdict(dict)
        cls.__handlers = []
        cls.__handler_cache = {}

    @classmethod
    def close(cls):
        pass

    @classmethod
    def _get_instance(cls, handler_class: type, agent, io) -> ResourceHandler:
        new_instance = handler_class(agent, io)
        return new_instance

    @classmethod
    def get_provider(cls, cache, agent, resource) -> ResourceHandler:
        """
            Return a provider to handle the given resource
        """
        resource_id = resource.id
        resource_type = resource_id.entity_type
        try:
            io = get_io(cache, agent.uri, resource_id.version)
        except Exception:
            LOGGER.exception("Exception raised during creation of IO for uri %s", agent.uri)
            raise Exception("No handler available for %s (no io available)" % resource_id)

        if io is None:
            # Skip this resource
            raise Exception("No handler available for %s (no io available)" % resource_id)

        available = []
        if resource_type in cls.__command_functions:
            for handlr in cls.__command_functions[resource_type].values():
                h = cls._get_instance(handlr, agent, io)
                if h.available(resource):
                    available.append(h)
                else:
                    h.close()

        if len(available) > 1:
            for h in available:
                h.close()

            io.close()
            raise Exception("More than one handler selected for resource %s" % resource.id)

        elif len(available) == 1:
            return available[0]

        raise Exception("No resource handler registered for resource of type %s" % resource_type)

    @classmethod
    def add_provider(cls, resource: str, name: str, provider):
        """
            Register a new provider

            :param resource: the name of the resource this handler applies to
            :param name: the name of the handler itself
            :param provider: the handler function
        """
        if resource in cls.__command_functions and name in cls.__command_functions[resource]:
            del cls.__command_functions[resource][name]

        cls.__command_functions[resource][name] = provider

    @classmethod
    def get_providers(cls) -> typing.Iterator[Tuple[str, typing.Type["ResourceHandler"]]]:
        """ Return an iterator over resource type, handler definition
        """
        for resource_type, handler_map in cls.__command_functions.items():
            for handle_name, handler_class in handler_map.items():
                yield (resource_type, handler_class)

    @classmethod
    def get_provider_class(cls, resource_type, name):
        """
            Return the class of the handler for the given type and with the given name
        """
        if resource_type not in cls.__command_functions:
            return None

        if name not in cls.__command_functions[resource_type]:
            return None

        return cls.__command_functions[resource_type][name]


class HandlerNotAvailableException(Exception):
    """
        This exception is thrown when a resource handler cannot perform its job. For example, the admin interface
        it connects to is not available.
    """



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""


class CLIException(Exception):

    def __init__(self, exitcode, *args, **kwargs):
        self.exitcode = exitcode
        super(CLIException, self).__init__(*args, **kwargs)


class Commander(object):
    """
        This class handles commands
    """
    __command_functions = {}

    @classmethod
    def add(cls, name, function, help_msg, parser_config, require_project=False, aliases=[]):
        """
            Add a new export function
        """
        if name in cls.__command_functions:
            raise Exception("Command %s already registered" % name)

        cls.__command_functions[name] = {
            "function": function,
            "help": help_msg,
            "parser_config": parser_config,
            "require_project": require_project,
            "aliases": aliases
        }

    config = None

    @classmethod
    def reset(cls):
        """
            Return a list of commands
        """
        cls.__command_functions = {}

    @classmethod
    def commands(cls):
        """
            Return a list of commands
        """
        return cls.__command_functions


class command(object):  # noqa: N801
    """
        A decorator that registers an export function
    """

    def __init__(self, name, help_msg, parser_config=None, require_project=False, aliases=[]):
        self.name = name
        self.help = help_msg
        self.require_project = require_project
        self.parser_config = parser_config
        self.aliases = aliases

    def __call__(self, function):
        """
            The wrapping
        """
        Commander.add(self.name, function, self.help, self.parser_config, self.require_project, self.aliases)
        return function



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import unittest
from time import sleep

from inmanta.agent.handler import cache
from inmanta.agent.cache import AgentCache
from inmanta.resources import resource, Resource, Id
import pytest
from _pytest.fixtures import fixture
from threading import Lock, Thread


@fixture(autouse=True)
def my_resource():
    @resource("test::Resource", agent="agent", id_attribute="key")
    class MyResource(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged")


class CacheTests(unittest.TestCase):

    def test_base(self):
        cache = AgentCache()
        value = "test too"
        cache.cache_value("test", value)
        assert value == cache.find("test")

    def test_timout(self):
        cache = AgentCache()
        value = "test too"
        cache.cache_value("test", value, timeout=0.1)
        cache.cache_value("test2", value)

        assert value == cache.find("test")
        sleep(1)
        try:
            assert value == cache.find("test")
            raise AssertionError("Should get exception")
        except KeyError:
            pass

        assert value == cache.find("test2")

    def test_base_fail(self):
        cache = AgentCache()
        value = "test too"
        with pytest.raises(KeyError):
            assert value == cache.find("test")

    def test_resource(self):
        cache = AgentCache()
        value = "test too"
        resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
        cache.cache_value("test", value, resource=resource)
        assert value == cache.find("test", resource=resource)

    def test_resource_fail(self):
        cache = AgentCache()
        value = "test too"
        resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
        cache.cache_value("test", value, resource=resource)

        with pytest.raises(KeyError):
            assert value == cache.find("test")

    def test_version_closed(self):
        cache = AgentCache()
        value = "test too"
        version = 200
        with pytest.raises(Exception):
            cache.cache_value("test", value, version=version)
            assert value == cache.find("test", version=version)

    def test_version(self):
        cache = AgentCache()
        value = "test too"
        version = 200
        cache.open_version(version)
        cache.cache_value("test", value, version=version)
        assert value == cache.find("test", version=version)

    def test_version_close(self):
        cache = AgentCache()
        value = "test too"
        version = 200
        cache.open_version(version)
        cache.cache_value("test", value, version=version)
        cache.cache_value("test0", value, version=version)
        cache.cache_value("test4", value, version=version)
        resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
        cache.cache_value("testx", value, resource=resource)
        assert value == cache.find("test", version=version)
        assert value == cache.find("testx", resource=resource)
        cache.close_version(version)
        assert value, cache.find("testx", resource=resource)
        try:
            assert value == cache.find("test", version=version)
            raise AssertionError("Should get exception")
        except KeyError:
            pass

    def test_multi_threaded(self):
        class Spy(object):

            def __init__(self):
                self.created = 0
                self.deleted = 0
                self.lock = Lock()

            def create(self):
                with self.lock:
                    self.created += 1
                return self

            def delete(self):
                self.deleted += 1

        cache = AgentCache()
        version = 200

        cache.open_version(version)

        alpha = Spy()
        beta = Spy()
        alpha.lock.acquire()

        t1 = Thread(target=lambda: cache.get_or_else(
            "test", lambda version: alpha.create(), version=version, call_on_delete=lambda x: x.delete()))
        t2 = Thread(target=lambda: cache.get_or_else(
            "test", lambda version: beta.create(), version=version, call_on_delete=lambda x: x.delete()))

        t1.start()
        t2.start()

        alpha.lock.release()

        t1.join()
        t2.join()

        assert alpha.created + beta.created == 1
        assert alpha.deleted == 0
        assert beta.deleted == 0

        cache.close_version(version)

        assert alpha.created + beta.created == 1
        assert alpha.deleted == alpha.created
        assert beta.deleted == beta.created

    def test_timout_and_version(self):
        cache = AgentCache()
        version = 200

        cache.open_version(version)
        value = "test too"
        cache.cache_value("test", value, version=version, timeout=0.3)
        cache.cache_value("testx", value)

        assert value == cache.find("test", version=version)
        assert value == cache.find("testx")

        sleep(1)
        assert value == cache.find("testx")

        cache.close_version(version)
        assert value == cache.find("testx")

        with pytest.raises(KeyError):
            cache.find("test", version=version)
        assert value == cache.find("testx")

    def test_version_and_timout(self):
        cache = AgentCache()
        version = 200

        cache.open_version(version)
        value = "test too"
        cache.cache_value("test", value, version=version, timeout=0.3)
        cache.cache_value("testx", value)

        assert value == cache.find("test", version=version)
        assert value == cache.find("testx")

        cache.close_version(version)
        assert value == cache.find("testx")

        sleep(1)
        assert value == cache.find("testx")

        with pytest.raises(KeyError):
            cache.find("test", version=version)

    def test_version_fail(self):
        cache = AgentCache()
        value = "test too"
        version = 200
        cache.open_version(version)
        cache.cache_value("test", value, version=version)

        with pytest.raises(KeyError):
            assert value == cache.find("test")

    def test_resource_and_version(self):
        cache = AgentCache()
        value = "test too"
        resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
        version = 200
        cache.open_version(version)
        cache.cache_value("test", value, resource=resource, version=version)
        assert value == cache.find("test", resource=resource, version=version)

    def test_get_or_else(self):
        called = []

        def creator(param, resource, version):

            called.append("x")
            return param

        cache = AgentCache()
        value = "test too"
        value2 = "test too x"
        resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
        resourcev2 = Id("test::Resource", "test", "key", "test", 200).get_instance()
        assert 200 == resourcev2.id.version
        version = 200
        cache.open_version(version)
        assert value == cache.get_or_else("test", creator, resource=resource, version=version, param=value)
        assert value == cache.get_or_else("test", creator, resource=resource, version=version, param=value)
        assert len(called) == 1
        assert value == cache.get_or_else("test", creator, resource=resourcev2, version=version, param=value)
        assert len(called) == 1
        assert value2 == cache.get_or_else("test", creator, resource=resource, version=version, param=value2)

    def test_get_or_else_none(self):
        called = []

        def creator(param, resource, version):
            called.append("x")
            return param

        class Sequencer(object):

            def __init__(self, sequence):
                self.seq = sequence
                self.count = 0

            def __call__(self, **kwargs):
                out = self.seq[self.count]
                self.count += 1
                return out

        cache = AgentCache()
        value = "test too"
        resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
        version = 100
        cache.open_version(version)
        assert None is cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=None)
        assert len(called) == 1
        assert None is cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=None)
        assert len(called) == 2
        assert value == cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=value)
        assert value == cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=value)
        assert len(called) == 3

        seq = Sequencer([None, None, "A"])
        assert None is cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
        assert seq.count == 1
        assert None is cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
        assert seq.count == 2
        assert "A" == cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
        assert seq.count == 3
        assert "A" == cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
        assert seq.count == 3
        assert "A" == cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
        assert seq.count == 3

    def test_decorator(self):

        xcache = AgentCache()

        class DT(object):

            def __init__(self, cache):
                self.cache = cache
                self.count = 0
                self.c2 = 0

            @cache
            def test_method(self):
                self.count += 1
                return "x"

            @cache
            def test_method_2(self, version):
                self.count += 1
                return "x2"

            @cache(cacheNone=False)
            def test_method_3(self):
                self.c2 += 1
                if self.c2 < 2:
                    return None
                else:
                    return "X"

        test = DT(xcache)
        assert "x" == test.test_method()
        assert "x" == test.test_method()
        assert "x" == test.test_method()
        assert 1 == test.count

        xcache.open_version(1)
        xcache.open_version(2)
        assert "x2" == test.test_method_2(version=1)
        assert "x2" == test.test_method_2(version=1)
        assert 2 == test.count
        assert "x2" == test.test_method_2(version=2)
        assert 3 == test.count
        xcache.close_version(1)
        xcache.open_version(1)
        assert "x2" == test.test_method_2(version=1)
        assert "x2" == test.test_method_2(version=1)
        assert 4 == test.count

        assert None is test.test_method_3()
        assert 1 == test.c2
        assert "X" == test.test_method_3()
        assert 2 == test.c2
        assert "X" == test.test_method_3()
        assert 2 == test.c2



import pytest
import tornado
from tornado.web import url

from tornado.httpserver import HTTPServer
from inmanta.reporter import InfluxReporter, AsyncReporter
from pyformance import timer
import asyncio
import re


class QueryMockHandler(tornado.web.RequestHandler):
    def initialize(self, parent):
        self.parent = parent

    def get(self, *args, **kwargs):
        self.parent.querycount += 1
        try:
            assert self.request.query_arguments["q"] == [b"CREATE DATABASE metrics"]
        except Exception as e:
            # carry over  failures
            self.parent.failure = e


influxlineprotocol = re.compile(r"\w+(,\w+=\w+)* (\w+=[\d.e+-]*)(,\w+=[\d.e+-]*)* \d+")


class WriteMockHandler(tornado.web.RequestHandler):
    def initialize(self, parent):
        self.parent = parent

    def post(self, *args, **kwargs):
        self.parent.writecount += 1
        try:
            for line in self.request.body.decode().split("\n"):
                print(line)
                assert influxlineprotocol.match(line)
                self.parent.lines.append(line)
        except Exception as e:
            # carry over  failures
            self.parent.failure = e


class InfluxdbMock(object):
    def __init__(self, socket):
        self.querycount = 0
        self.writecount = 0
        self.failure = None
        self.lines = []

        self.app = tornado.web.Application(
            [
                url(r"/query", QueryMockHandler, kwargs={"parent": self}),
                url(r"/write", WriteMockHandler, kwargs={"parent": self}),
            ]
        )

        self.server = HTTPServer(self.app)
        self.server.add_sockets([socket])
        _addr, port = socket.getsockname()
        self.port = port


class MockReporter(AsyncReporter):
    def __init__(self, interval):
        super().__init__(None, interval)
        self.waiter = asyncio.locks.Semaphore(0)
        self.in_count = 0
        self.count = 0

    async def report_now(self, registry=None, timestamp=None) -> None:
        self.in_count += 1
        await self.waiter.acquire()
        self.count += 1


@pytest.fixture
def influxdb(event_loop, free_socket):
    ifl = InfluxdbMock(free_socket)
    yield ifl
    ifl.server.stop()


@pytest.mark.asyncio
async def test_influxdb(influxdb):
    rep = InfluxReporter(
        port=influxdb.port, tags={"mark": "X"}, autocreate_database=True
    )
    with timer("test").time():
        pass

    with timer("test2").time():
        pass

    await rep.report_now()

    assert influxdb.querycount == 1
    assert influxdb.writecount == 1

    if influxdb.failure:
        raise influxdb.failure

    for line in influxdb.lines:
        assert "mark=X" in line


@pytest.mark.asyncio
async def test_timing():
    # Attempt to deploy every 0.01 seconds,
    # Deploy hangs on a semaphore, so test case can control progress
    mr = MockReporter(0.01)
    mr.start()

    # no deploy done
    assert mr.count == 0
    # wait for deploy
    await asyncio.sleep(0.01)
    # one is waiting
    assert mr.in_count == 1
    # release
    mr.waiter.release()
    # wait 0 to allow reporter to make progress
    await asyncio.sleep(0.0)
    # one deploy done
    assert mr.count == 1

    # allow to run free to test timing
    for i in range(5):
        mr.waiter.release()

    # wait 0 to allow reporter to make progress
    await asyncio.sleep(0)
    # how many are we now?
    base = mr.count
    # should be sufficiently far of the lock
    # otherwise test always succeeds
    assert base < 4
    # wait for one more
    await asyncio.sleep(0.01)
    assert mr.count == base + 1
    mr.stop()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import logging
import uuid
import datetime
import json
from collections import defaultdict


from inmanta import protocol
from inmanta.config import Config, cmdline_rest_transport
from inmanta.const import AgentTriggerMethod, TIME_ISOFMT
from inmanta.resources import Id
import click
import texttable
from time import sleep


from typing import Optional, cast, Dict, List, Callable

from inmanta.types import JsonType


class Client(object):
    log = logging.getLogger(__name__)

    def __init__(self, host: Optional[str], port: Optional[int]) -> None:
        if host is None:
            self.host = cast(str, cmdline_rest_transport.host.get())
        else:
            self.host = host
            Config.set("cmdline_rest_transport", "host", host)

        if port is None:
            self.port = cast(int, cmdline_rest_transport.port.get())
        else:
            self.port = port
            Config.set("cmdline_rest_transport", "port", str(port))

        self._client = protocol.SyncClient("cmdline")

    def do_request(
        self, method_name: str, key_name: Optional[str]=None, arguments: JsonType={}, allow_none: bool=False
    ) -> Optional[JsonType]:
        """
            Do a request and return the response
        """
        self.log.debug(
            "Calling method %s on server %s:%s with arguments %s", method_name, self.host, self.host, arguments
        )

        if not hasattr(self._client, method_name):
            raise Exception("API call %s is not available." % method_name)

        method: Callable[..., protocol.Result] = getattr(self._client, method_name)
        result = method(**arguments)

        if result is None:
            raise Exception("Failed to call server.")

        self.log.debug("Got response code %s and data: %s", result.code, result.result)

        if result.code == 200:
            if key_name is None:
                return result.result

            if key_name in result.result:
                return result.result[key_name]

            raise Exception("Expected %s in the response of %s." % (key_name, method_name))
        elif result.code == 404:
            if not allow_none:
                raise Exception("Requested %s not found on server" % key_name)
            return None

        else:
            msg = ": "
            if result.result is not None and "message" in result.result:
                msg += result.result["message"]

            raise Exception(("An error occurred while requesting %s" % key_name) + msg)

    def get_list(self, method_name: str, key_name: Optional[str]=None, arguments: JsonType={}) -> List[Dict[str, str]]:
        """
            Same as do request, but return type is a list of dicts
        """
        return cast(List[Dict[str, str]], self.do_request(method_name, key_name, arguments, False))

    def get_dict(self, method_name: str, key_name: Optional[str] = None, arguments: JsonType = {}) -> Dict[str, str]:
        """
            Same as do request, but return type is a list of dicts
        """
        return cast(Dict[str, str], self.do_request(method_name, key_name, arguments, False))

    def to_project_id(self, ref: str) -> uuid.UUID:
        """
            Convert ref to a uuid
        """
        try:
            project_id = uuid.UUID(ref)
        except ValueError:
            # try to resolve the id as project name
            projects = self.get_list("list_projects", "projects")

            id_list = []
            for project in projects:
                if ref == project["name"]:
                    id_list.append(project["id"])

            if len(id_list) == 0:
                raise Exception("Unable to find a project with the given id or name")

            elif len(id_list) > 1:
                raise Exception("Found multiple projects with %s name, please use the ID." % ref)

            else:
                project_id = uuid.UUID(id_list[0])

        return project_id

    def to_environment_id(self, ref: str, project_id: uuid.UUID=None) -> uuid.UUID:
        """
            Convert ref to an env uuid, optionally scoped to a project
        """
        try:
            env_id = uuid.UUID(ref)
        except ValueError:
            # try to resolve the id as project name
            envs = self.get_list("list_environments", "environments")

            id_list = []
            for env in envs:
                if ref == env["name"]:
                    if project_id is None or project_id == env["project_id"]:
                        id_list.append(env["id"])

            if len(id_list) == 0:
                raise Exception("Unable to find an environment with the given id or name")

            elif len(id_list) > 1:
                raise Exception("Found multiple environment with %s name, please use the ID." % ref)

            else:
                env_id = uuid.UUID(id_list[0])

        return env_id

    def to_form_id(self, ref: str, environment: uuid.UUID) -> uuid.UUID:
        """
            Convert ref to a form uuid
        """
        try:
            env_id = uuid.UUID(ref)
        except ValueError:
            # try to resolve the id as project name
            forms = self.get_list("list_forms", "forms", arguments=dict(tid=environment))

            id_list = []
            for form in forms:
                if ref == form["form_type"]:
                    id_list.append(form["form_id"])

            if len(id_list) == 0:
                raise Exception("Unable to find a form with the given id or name")

            elif len(id_list) > 1:
                raise Exception("Found multiple forms with %s name, please use the ID." % ref)

            else:
                env_id = uuid.UUID(id_list[0])

        return env_id


def print_table(header: List[str], rows: List[List[str]], data_type: List[str]=None) -> None:
    width, _ = click.get_terminal_size()

    table = texttable.Texttable(max_width=width)
    table.set_deco(texttable.Texttable.HEADER | texttable.Texttable.BORDER | texttable.Texttable.VLINES)
    if data_type is not None:
        table.set_cols_dtype(data_type)
    table.header(header)
    for row in rows:
        table.add_row(row)
    click.echo(table.draw())


@click.group()
@click.option("--host", help="The server hostname to connect to")
@click.option("--port", help="The server port to connect to")
@click.pass_context
def cmd(ctx: click.Context, host: str, port: int) -> None:
    ctx.obj = Client(host, port)


@cmd.group("project")
@click.pass_context
def project(ctx: click.Context) -> None:
    pass


@project.command(name="list")
@click.pass_obj
def project_list(client: Client) -> None:
    projects = client.get_list("list_projects", "projects")

    if len(projects) > 0:
        print_table(['ID', 'Name'], [[n['id'], n['name']] for n in projects])

    else:
        click.echo("No projects defined.", err=True)


@project.command(name="show")
@click.argument("project")  # , help="The the id or name of the project to show")
@click.pass_obj
def project_show(client: Client, project: str) -> None:
    project_id = client.to_project_id(project)
    project_data = client.get_dict("get_project", "project", dict(id=project_id))

    print_table(["Name", "Value"], [["ID", project_data["id"]], ["Name", project_data["name"]]])


@project.command(name="create")
@click.option("--name", "-n", help="The name of the new project", required=True)
@click.pass_obj
def project_create(client: Client, name: str) -> None:
    project = client.get_dict("create_project", "project", {"name": name})
    print_table(["Name", "Value"], [["ID", project["id"]], ["Name", project["name"]]])


@project.command(name="modify")
@click.option("--name", "-n", help="The new name of the project", required=True)
@click.argument("project")  # , help="The id of the project to modify")
@click.pass_obj
def project_modify(client: Client, name: str, project: str) -> None:
    project_id = client.to_project_id(project)
    project_data = client.get_dict("modify_project", "project", dict(id=project_id, name=name))
    print_table(["Name", "Value"], [["ID", project_data["id"]], ["Name", project_data["name"]]])


@project.command(name="delete")
@click.argument("project")  # , help="The id of the project to modify")
@click.pass_obj
def project_delete(client: Client, project: str) -> None:
    project_id = client.to_project_id(project)
    client.do_request("delete_project", arguments={"id": project_id})
    click.echo("Project successfully deleted")


@cmd.group("environment")
@click.pass_context
def environment(ctx: click.Context) -> None:
    pass


@environment.command(name="create")
@click.option("--name", "-n", help="The name of the new environment", required=True)
@click.option("--project", "-p", help="The id of the project this environment belongs to", required=True)
@click.option("--repo-url", "-r", required=False, default="",
              help="The url of the repository that contains the configuration model")
@click.option("--branch", "-b", required=False, default="master",
              help="The branch in the repository that contains the configuration model")
@click.option("--save", "-s", default=False, is_flag=True,
              help="Save the ID of the environment and the server to the .inmanta config file")
@click.pass_obj
def environment_create(client: Client, name: str, project: str, repo_url: str, branch: str, save: bool) -> None:
    project_id = client.to_project_id(project)
    env = client.get_dict(
        "create_environment",
        "environment",
        dict(project_id=project_id, name=name, repository=repo_url, branch=branch)
    )
    project_data = client.get_dict("get_project", "project", {"id": project_id})

    if save:
        cfg = """
[config]
fact-expire = 1800
environment=%(env)s

[compiler_rest_transport]
host=%(host)s
port=%(port)s

[cmdline_rest_transport]
host=%(host)s
port=%(port)s
""" % {"env": env["id"], "host": client.host, "port": client.port}
        if os.path.exists(".inmanta"):
            click.echo(".inmanta exits, not writing config", err=True)
        else:
            with open(".inmanta", 'w') as f:
                f.write(cfg)

    print_table(
        ['Environment ID', 'Environment name', 'Project ID', 'Project name'],
        [[env["id"], env["name"], project_data["id"], project_data["name"]]]
    )


@environment.command(name="list")
@click.pass_obj
def environment_list(client: Client) -> None:
    environments = client.get_list("list_environments", "environments")

    data = []
    for env in environments:
        prj = client.get_dict("get_project", "project", dict(id=env["project"]))
        prj_name = prj['name']
        data.append([prj_name, env['project'], env['name'], env['id']])

    if len(data) > 0:
        print_table(['Project name', 'Project ID', 'Environment', 'Environment ID'], data)
    else:
        click.echo("No environment defined.")


@environment.command(name="show")
@click.argument("environment")
@click.pass_obj
def environment_show(client: Client, environment: str) -> None:
    env = client.get_dict("get_environment", "environment", dict(id=client.to_environment_id(environment)))
    print_table(
        ['ID', 'Name', 'Repository URL', 'Branch Name'],
        [[env["id"], env["name"], env["repo_url"], env["repo_branch"]]],
    )


@environment.command(name="modify")
@click.option("--name", "-n", help="The name of the new environment", required=True)
@click.option("--repo-url", "-r", required=False, default="",
              help="The url of the repository that contains the configuration model")
@click.option("--branch", "-b", required=False, default="master",
              help="The branch in the repository that contains the configuration model")
@click.argument("environment")
@click.pass_obj
def environment_modify(client: Client, environment: str, name: str, repo_url: str, branch: str) -> None:
    env = client.get_dict(
        "modify_environment",
        "environment",
        dict(id=client.to_environment_id(environment), name=name, repository=repo_url, branch=branch)
    )

    print_table(
        ['ID', 'Name', 'Repository URL', 'Branch Name'],
        [[env["id"], env["name"], env["repo_url"], env["repo_branch"]]]
    )


@environment.command(name="delete")
@click.argument("environment")
@click.pass_obj
def environment_delete(client: Client, environment: str) -> None:
    env_id = client.to_environment_id(environment)
    client.do_request("delete_environment", arguments=dict(id=env_id))
    click.echo("Environment successfully deleted")


@environment.group("setting")
@click.pass_context
def env_setting(ctx: click.Context) -> None:
    pass


@env_setting.command(name="list")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def env_setting_list(client: Client, environment: str) -> None:
    tid = client.to_environment_id(environment)
    settings = cast(Dict[str, Dict[str, str]], client.do_request("list_settings", arguments=dict(tid=tid)))

    table_body = []
    for key in sorted(settings["metadata"].keys()):
        meta = cast(Dict[str, str], settings["metadata"][key])
        value = ""
        if key in settings["settings"]:
            value = str(settings["settings"][key])

        default_value = ""
        if "default" in meta:
            default_value = str(meta["default"])

        table_body.append([key, value, default_value, meta["type"], meta["doc"]])

    click.echo("Settings for environment %s" % tid)
    print_table(["Key", "Value", "Default value", "Type", "Help"], table_body)


@env_setting.command(name="set")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--key", "-k", help="The key to set", required=True)
@click.option("--value", "-o", help="The value to set", required=True)
@click.pass_obj
def env_setting_set(client: Client, environment: str, key: str, value: str) -> None:
    tid = client.to_environment_id(environment)
    client.do_request("set_setting", arguments=dict(tid=tid, id=key, value=value))


@env_setting.command(name="get")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--key", "-k", help="The key to get", required=True)
@click.pass_obj
def env_setting_get(client: Client, environment: str, key: str) -> None:
    tid = client.to_environment_id(environment)
    value = client.get_dict("get_setting", arguments=dict(tid=tid, id=key))
    click.echo(value["value"])


@env_setting.command(name="delete")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--key", "-k", help="The key to delete", required=True)
@click.pass_obj
def env_setting_del(client: Client, environment: str, key: str) -> None:
    tid = client.to_environment_id(environment)
    client.do_request("delete_setting", arguments=dict(tid=tid, id=key))


@cmd.group("agent")
@click.pass_context
def agent(ctx: click.Context) -> None:
    pass


@agent.command(name="list")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def agent_list(client: Client, environment: str) -> None:
    env_id = client.to_environment_id(environment)
    agents = client.get_list("list_agents", key_name="agents", arguments=dict(tid=env_id))
    data = []
    for agent in agents:
        data.append([agent["name"], agent["environment"], agent["last_failover"]])

    print_table(['Agent', 'Environment', 'Last fail over'], data)


@cmd.group("version")
@click.pass_context
def version(ctx: click.Context) -> None:
    pass


@version.command(name="list")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def version_list(client: Client, environment: str) -> None:
    env_id = client.to_environment_id(environment)
    versions = client.get_list("list_versions", "versions", arguments=dict(tid=env_id))

    print_table(
        ['Created at', 'Version', 'Released', 'Deployed', '# Resources', '# Done', 'State'],
        [[x['date'], x['version'], x['released'], x['deployed'], x['total'], x['done'], x['result']] for x in versions],
        ["t", "t", "t", "t", "t", "t", "t"]
    )


@version.command(name="release")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--push", "-p", help="Push the version to the deployment agents", is_flag=True)
@click.option("--full", help="Make the agents execute a full deploy instead of an incremental deploy. "
                             "Should be used together with the --push option", is_flag=True)
@click.argument("version")
@click.pass_obj
def version_release(client: Client, environment: str, push: bool, full: bool, version: str) -> None:
    env_id = client.to_environment_id(environment)
    if push:
        trigger_method = AgentTriggerMethod.get_agent_trigger_method(full)
        x = client.get_dict("release_version", "model", dict(tid=env_id, id=version, push=True,
                                                             agent_trigger_method=trigger_method))
    else:
        x = client.get_dict("release_version", "model", dict(tid=env_id, id=version))

    print_table(
        ['Created at', 'Version', 'Released', 'Deployed', '# Resources', '# Done', 'State'],
        [[x['date'], x['version'], x['released'], x['deployed'], x['total'], x['done'], x['result']]]
    )


@cmd.group("param")
@click.pass_context
def param(ctx: click.Context) -> None:
    pass


@param.command(name="list")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def param_list(client: Client, environment: str) -> None:
    result = client.get_dict("list_params", arguments=dict(tid=client.to_environment_id(environment)))
    expire = int(result["expire"])
    now = datetime.datetime.strptime(result["now"], TIME_ISOFMT)
    when = now - datetime.timedelta(0, expire)

    data = []
    parameters = cast(List[Dict[str, str]], result["parameters"])
    for p in parameters:
        data.append(
            [
                p["resource_id"],
                p['name'],
                p['source'],
                p['updated'],
                str(float(datetime.datetime.strptime(p["updated"], TIME_ISOFMT) < when))
            ]
        )

    print_table(['Resource', 'Name', 'Source', 'Updated', 'Expired'], data)


@param.command(name="set")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--name", help="The name of the parameter", required=True)
@click.option("--value", help="The value of the parameter", required=True)
@click.pass_obj
def param_set(client: Client, environment: str, name: str, value: str) -> None:
    tid = client.to_environment_id(environment)
    # first fetch the parameter
    param_data = cast(
        Optional[Dict[str, str]],
        client.do_request("get_param", "parameter", dict(tid=tid, id=name, resource_id=""), allow_none=True)
    )

    param = {"source": "user", "metadata": {}} if param_data is None else param_data
    param_return = client.get_dict(
        "set_param",
        "parameter",
        dict(tid=tid, id=name, value=value, source=param["source"], resource_id="", metadata=param["metadata"])
    )

    print_table(
        ['Name', 'Value', 'Source', 'Updated'],
        [[param_return['name'], param_return['value'], param_return['source'], param_return['updated']]]
    )


@param.command(name="get")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--name", help="The name of the parameter", required=True)
@click.option("--resource", help="The resource id of the parameter")
@click.pass_obj
def param_get(client: Client, environment: str, name: str, resource: str) -> None:
    tid = client.to_environment_id(environment)

    if resource is None:
        resource = ""

    # first fetch the parameter
    param = client.get_dict("get_param", "parameter", dict(tid=tid, id=name, resource_id=resource))

    print_table(
        ['Name', 'Value', 'Source', 'Updated'],
        [[param['name'], param['value'], param['source'], param['updated']]]
    )


@version.command(name="report")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--version", "-i", help="The version to create a report from", required=True)
@click.option("-l", is_flag=True, help="Show a detailed version of the report")
@click.pass_obj
def version_report(client: Client, environment: str, version: str, l: bool) -> None:
    tid = client.to_environment_id(environment)
    result = client.do_request("get_version", arguments=dict(tid=tid, id=version, include_logs=True))

    agents: Dict[str, Dict[str, List[str]]] = defaultdict(lambda: defaultdict(lambda: []))
    for res in result["resources"]:
        if len(res["actions"]) > 0 or l:
            agents[res["agent"]][res["resource_type"]].append(res)

    for agent in sorted(agents.keys()):
        click.echo(click.style("Agent: %s" % agent, bold=True))
        click.echo("=" * 72)

        for t in sorted(agents[agent].keys()):
            parsed_resource_version_id = Id.parse_id(agents[agent][t][0]["resource_version_id"])
            click.echo(click.style("Resource type:", bold=True)
                       + "{type} ({attr})".format(type=t, attr=parsed_resource_version_id.attribute))
            click.echo("-" * 72)

            for res in agents[agent][t]:
                parsed_id = Id.parse_id(res["resource_version_id"])
                click.echo((click.style(parsed_id.attribute_value, bold=True) + " (#actions=%d)") % len(res["actions"]))
                # for dryrun show only the latest, for deploy all
                if not result["model"]["released"]:
                    if len(res["actions"]) > 0:
                        action = res["actions"][0]
                        click.echo("* last check: %s" % action["timestamp"])
                        click.echo("* result: %s" % ("error" if action["level"] != "INFO" else "success"))
                        if len(action["data"]) == 0:
                            click.echo("* no changes")
                        else:
                            click.echo("* changes:")
                            for field in sorted(action["data"].keys()):
                                values = action["data"][field]
                                if field == "hash":
                                    click.echo("  - content:")
                                    diff_value = client.do_request("diff", arguments=dict(a=values[0], b=values[1]))
                                    click.echo("    " + "    ".join(diff_value["diff"]))
                                else:
                                    click.echo("  - %s:" % field)
                                    click.echo("    " + click.style("from:", bold=True) + " %s" % values[0])
                                    click.echo("    " + click.style("to:", bold=True) + " %s" % values[1])

                                click.echo("")

                        click.echo("")
                else:
                    pass

            click.echo("")


@cmd.group("form")
@click.pass_context
def form(ctx: click.Context) -> None:
    pass


@form.command(name="list")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def form_list(client: Client, environment: str) -> None:
    result = client.get_list("list_forms", "forms", arguments=dict(tid=client.to_environment_id(environment)))

    data = []
    for p in result:
        data.append([p["form_type"], p['form_id']])

    print_table(['Form Type', 'Form ID'], data)


@form.command(name="show")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--form-type", "-t", help="Show details of this form", required=True)
@click.pass_obj
def form_show(client: Client, environment: str, form_type: str) -> None:
    result = client.get_dict("get_form", "form", arguments=dict(tid=client.to_environment_id(environment), id=form_type))
    values = []
    for k, v in result["fields"].items():
        if k in result["defaults"] and result["defaults"][k] != "":
            values.append([k, "type: %s, default: %s" % (v, result["defaults"][k])])
        else:
            values.append([k, "type: %s" % v])

    print_table(["Field", "Spec"], values)


@form.command(name="export")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--form-type", "-t", help="Show details of this form", required=True)
@click.pass_obj
def form_export(client: Client, environment: str, form_type: str) -> None:
    tid = client.to_environment_id(environment)
    form_def = client.get_dict("get_form", "form", arguments=dict(tid=tid, id=form_type))
    form_records = client.get_list(
        "list_records",
        "records",
        arguments=dict(tid=tid, form_type=form_type, include_record=True)
    )

    click.echo(json.dumps({"form_type": form_def, "records": form_records}))


@form.command(name="import")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--form-type", "-t", help="Show details of this form", required=True)
@click.option("--file", help="The json file with the record data", required=True)
@click.pass_obj
def form_import(client: Client, environment: str, form_type: str, file: str) -> None:
    tid = client.to_environment_id(environment)
    if not os.path.exists(file):
        raise Exception("%s file does not exist." % file)

    data = {}
    with open(file, "r") as fd:
        try:
            data = json.load(fd)
        except Exception as e:
            raise Exception("Unable to load records, invalid json") from e

    if "records" not in data:
        raise Exception("No records found in input file")

    form_type_def: Dict[str, str] = data["form_type"]
    if form_type != form_type_def["form_type"]:
        raise click.ClickException("Unable to load form data for %s into form %s" % (form_type_def["form_type"], form_type))

    records = cast(List[JsonType], data["records"])
    for record in records:
        if record["form"] == form_type:
            client.do_request("create_record", "record", arguments=dict(tid=tid, form_type=form_type, form=record["fields"]))


@cmd.group("record")
@click.pass_context
def record(ctx: click.Context) -> None:
    pass


@record.command(name="list")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--form-type", "-t", help="Show details of this form", required=True)
@click.option("--show-all", "-a", help="Show all fields", is_flag=True, default=False)
@click.pass_obj
def record_list(client: Client, environment: str, form_type: str, show_all: bool) -> None:
    tid = client.to_environment_id(environment)

    if not show_all:
        result = client.get_list("list_records", "records", arguments=dict(tid=tid, form_type=form_type))
        data = []
        for p in result:
            data.append([p["id"], p['changed']])

        print_table(['Record ID', 'Changed'], data)
    else:
        result = client.get_list("list_records", "records", arguments=dict(tid=tid, form_type=form_type, include_record=True))
        fields = []
        data = []
        for p in result:
            fields = p["fields"].keys()
            values = [p["id"], p['changed']]
            values.extend(p["fields"].values())
            data.append(values)

        allfields = ['Record ID', 'Changed']
        allfields.extend(fields)
        print_table(allfields, data)


@record.command(name="create")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--form-type", "-t", help="Create a record of this type.", required=True)
@click.option("--field", "-p", help="Field values", multiple=True, default=[])
@click.pass_obj
def record_create(client: Client, environment: str, form_type: str, field: List[str]) -> None:
    tid = client.to_environment_id(environment)

    fields = {}
    for f in field:
        parts = f.split("=")
        if len(parts) != 2:
            raise Exception("Argument %s should be in the key=value form." % f)

        fields[parts[0].strip()] = parts[1].strip()

    try:
        uuid.UUID(form_type)
        raise Exception("Form type should be the type string, not the uuid.")
    except ValueError:
        pass

    result = client.get_list("create_record", "record", arguments=dict(tid=tid, form_type=form_type, form=fields))

    values = []
    fields = cast(List[Dict[str, str]], result["fields"])
    for k in sorted(fields.keys()):
        values.append([k, fields[k]])

    print_table(["Field", "Value"], values)


@record.command(name="update")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--record", "-r", help="The id of the record to edit", required=True)
@click.option("--field", "-p", help="Field values", multiple=True, default=[])
@click.pass_obj
def record_update(client: Client, environment: str, record: str, field: List[str]) -> None:
    tid = client.to_environment_id(environment)

    fields = {}
    for f in field:
        parts = f.split("=")
        if len(parts) != 2:
            raise Exception("Argument %s should be in the key=value form." % f)

        fields[parts[0].strip()] = parts[1].strip()

    result = cast(
        Dict[str, Dict[str, str]],
        client.do_request("update_record", "record", arguments=dict(tid=tid, id=record, form=fields))
    )

    values = []
    for k in sorted(result["fields"].keys()):
        values.append([k, result["fields"][k]])

    print_table(["Field", "Value"], values)


@record.command(name="delete")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.argument("record")
@click.pass_obj
def record_delete(client: Client, environment: str, record: str) -> None:
    tid = client.to_environment_id(environment)
    try:
        record_id = uuid.UUID(record)
    except ValueError:
        raise Exception("The record id should be a valid UUID")

    client.do_request("delete_record", arguments=dict(tid=tid, id=record_id))


@cmd.command(name="monitor")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def monitor_deploy(client: Client, environment: str) -> None:
    tid = client.to_environment_id(environment)

    versions = cast(Dict[str, List[Dict[str, str]]], client.do_request("list_versions", arguments=dict(tid=tid)))
    allversion = versions["versions"]
    try:
        first: Dict[str, str] = next(version for version in allversion if version["result"] != "pending")
    except StopIteration:
        raise click.ClickException("Environment %s doesn't contain a released configuration model" % environment)

    total = int(first["total"])
    done = int(first["done"])
    last = done
    ident = int(first["version"])

    with click.progressbar(label="version:%d" % ident, length=total, show_pos=True, show_eta=False) as bar:
        bar.update(done)
        while done != total:
            if done != last:
                bar.update(done - last)
                last = done
            sleep(1)
            version = cast(
                Dict[str, Dict[str, int]],
                client.get_dict("get_version", arguments=dict(tid=tid, id=int(ident), limit=0))
            )
            done = version["model"]["done"]
        if done != last:
            bar.update(done - last)
            last = done

    click.echo("Complete: %s/%s" % (done, total))


@cmd.group("token")
@click.pass_context
def token(ctx: click.Context) -> None:
    pass


@token.command(name="create")
@click.option("--environment", "-e", help="The environment to use.", required=True)
@click.option("--api", is_flag=True, help="Add client_type api to the token.")
@click.option("--compiler", is_flag=True, help="Add client_type compiler to the token.")
@click.option("--agent", is_flag=True, help="Add client_type agent to the token.")
@click.pass_obj
def create_token(client: Client, environment: str, api: str, compiler: str, agent: str) -> None:
    tid = client.to_environment_id(environment)

    client_types = []
    if api:
        client_types.append("api")

    if compiler:
        client_types.append("compiler")

    if agent:
        client_types.append("agent")

    token = cast(str, client.do_request("create_token", key_name="token", arguments=dict(tid=tid, client_types=client_types)))

    click.echo("Token: " + token)


@token.command(name="bootstrap")
@click.pass_obj
def bootstrap_token(client: Client) -> None:
    """
        Generate a bootstrap token that provides access to everything. This token is only valid for 3600 seconds.
    """
    click.echo("Token: " + protocol.encode_token(["api", "compiler", "agent"], expire=3600))


def main() -> None:
    Config.load_config()
    cmd()


if __name__ == '__main__':
    main()
